<?php

declare(strict_types=1);

namespace hood\vk\api\methods;

use Exception;

use hood\accounts\vk;
use hood\vk\robots\robot;
use hood\vk\api\data;
use hood\vk\robots\group;

/**
 * Сообщение
 *
 * @method public static function put(RobotAbstract $from, int $to, string $message, int $mode = 2) Отправить сообщение
 *
 * @see https://vk.com/dev/messages.send
 * @see https://vk.com/dev/messages.getById
 *
 * @package hood\vk\api
 * @author Arsen Mirzaev Tatyano-Muradovich <red@hood.su>
 *
 * @todo Доработать строгий режим отправки: проверку сообщения в беседе (не имеет ID сообщений)
 */
final class messages extends method
{
    /**
     * @param int $mode Режим отправки
     */
    protected int $mode = 1;

    /**
     * @param string|null $text Текст
     */

    /**
     *  @var data Вложения
     */
    protected data $data;

    /**
     * @var array Пересылаемые сообщения
     */
    protected array $forwardMessages = [];

    /**
     * @var int $ReplyMessage Ответное сообщение
     */
    protected int $ReplyMessage;

    /**
     * Создать сообщение
     *
     * @param robot $robot Робот
     */
    public function __construct(
        protected robot $robot,
        protected string|null $text = null
    ) {
    }

    /**
     * Записать текст
     *
     * @param string $text Текст
     *
     * @return self
     */
    public function text(string $text): self
    {
        // Записать текст
        if (!isset($this->text)) {
            $this->text = $text;
        } else {
            $this->text .= $text;
        }

        return $this;
    }

    /**
     * Записать пересылаемые сообщения
     * 
     * @param $messageIds id пересылаемых сообщений
     * 
     * @return self
     */
    public function forward(...$messageIds): self
    {
        // Запись пересылаемых сообщений
        $this->forwardMessages = array_merge($this->forwardMessages, $messageIds);

        return $this;
    }

    /**
     * Записать ответное сообщение
     * 
     * @param $messageId id ответного сообщения
     * 
     * @return self
     */
    public function reply(int $messageId): self
    {
        // Запись ответного сообщения
        $this->ReplyMessage = $messageId;

        return $this;
    }

    /**
     * Записать фото
     *
     * @param $img Фото
     *
     * @return self
     */
    public function image(...$pathPhoto): self
    {
        // Перебор фото
        foreach ($pathPhoto as $photo) {

            // Загрузить фото
            $id = $this->robot->photo()->getPhoto($photo);

            // Записать к вложениям
            $this->attachments($id);
        }

        return $this;
    }

    /**
     * Записать вложение
     *
     * @param $attachments Вложения
     * 
     * @return self
     */
    public function attachments(string ...$attachments): self
    {
        if (isset($this->data)) {
            // Если вложения инициализированны

            // Записать вложение
            foreach ($attachments as $attachment) {
                $this->data->addData($attachment);
            }
        } else {
            // Если вложения не инициализированны

            // Инициализация вложений
            $this->data = new data($this->robot);

            // Записать вложение
            foreach ($attachments as $attachment) {
                $this->data->addData($attachment);
            }
        }

        return $this;
    }

    /**
     * Отправить сообщение
     *
     * @param int|string|array $destination Получатель
     *
     * @see https://vk.com/dev/messages.send
     *
     * @return array Ответ сервера
     */
    public function send(int|string|array $destination): mixed
    {
        // Идентификатор
        $random_id = time();

        if ($this->mode = 1) {
            // Перемножение (по умолчанию)
            $random_id *= rand();
        }

        // Реиницилазиция
        $this->robot->api->reinit();

        // Цель отправки
        $this->robot->api->chooseDestination($destination);

        // Идентификатор сообщения
        $this->robot->api['random_id'] = $random_id;

        // Текст
        $this->robot->api['message'] = $this->text;

        // Пересылаемые сообщения
        if (!empty($this->forwardMessages)) {

            $this->robot->api['forward_messages'] = implode(',', $this->forwardMessages);
        }

        // Ответные сообщения
        if (isset($this->ReplyMessage)) {

            $this->robot->api['reply_to'] = $this->ReplyMessage;
        }

        // Вложения
        if ( isset($this->data) && $this->__get('data') !== []) { // !empty($this->data->data) почемуто не работает

            $this->robot->api['attachment'] = implode(',', $this->__get('data'));
        }

        // Запрос
        $request = json_decode($this->robot->browser->request('POST', 'messages.send', ['form_params' => $this->robot->api->settings])->getBody()->getContents());

        // Если в ответе ошибка
        if (isset($request->error)) {
            throw new Exception('Вконтакте: ' . $request->error->error_msg, $request->error->error_code);
        }

        if ($this->mode >= 2) {
            // Если установлен режим 2 (усиленная проверка отправленного сообщения)

            if (!empty($request["response"])) {
                // Ответ получен

                // Реиницилазиция
                $this->robot->api->reinit();

                // Запрашиваемые сообщения
                $this->robot->api['message_ids'] = $request["response"];

                // Запрос
                if ($this->robot->browser->request('POST', 'messages.getById', ['form_params' => $this->robot->api->settings])['response']['count'] === 0) {
                    // Сообщения не существует
                    //!!!!!!!!!!!!!!!!!!!!!!!!!!!
                    //!!!!!!!!!!!!!!!!!!!!!!!!!!
                    //!!!!!!!!!!!!!!!!!!!!!!!!!

                    // Повторная отправка
                    $this->send($destination);
                }
            } else {
                // Что-то придумать :)
                // наверно throw new exception($request->error->error_msg)
                // А стоп. Ничего не нужно, просто возвращаем объект с ошибкой, где всё будет написано
                // Костя ты тупой?
            }
        }

        return $request->response;
    }

    /**
     * Записать свойство
     *
     * @param string $name Название
     * @param mixed $value Значение
     *
     * @return void
     */
    public function __set(string $name, mixed $value): void
    {
        match ($name) {
            'data' => !isset($this->data) ? $this->data->addData($value) : throw new Exception('Вложения не инициализированны'),
            default => throw new Exception('Свойство не найдено: ' . $name)
        };
    }

    /**
     * Прочитать свойство
     *
     * @param string $name Название
     *
     * @return mixed
     */
    public function __get(string $name): mixed
    {
        return match ($name) {
            'text' => $this->text ?? throw new Exception('Текст не инициализирован'),
            'data' => isset($this->data) ? $this->data->data : throw new Exception('Вложения не инициализированны'),
            'forwardMessages' => !empty($this->forwardMessages) ? $this->forwardMessages : throw new Exception('Пересылаемые сообщения не инициализированны'),
            'ReplyMessage' => isset($this->ReplyMessage) ? $this->ReplyMessage : throw new Exception('Ответное сообщение не инициализированно'),
            default => throw new Exception('Свойство не найдено: ' . $name)
        };
    }

    /**
     * Получить информацию о сообщении по id
     * 
     * @param string $message_ids Идентификатор сообщения
     *
     * @return object Информация о сообщении
     */

    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    // доделать

    public function getById(int $message_ids): object
    {
        // Реиницилазиция
        $this->robot->api->reinit();

        // Идентификаторы сообщений. Максимум 100 идентификаторов
        $this->robot->api['message_ids'] = $message_ids;


        if ($this->robot instanceof group) {
            // Если это группа

            // Идентификатор группы
            isset($this->robot->id) ? $this->robot->api['group_id'] = $this->robot->id : throw new Exception('Необходимо указать идентификатор сообщества');
        }

        // Запрос
        $request = json_decode(($this->robot->browser->request('POST', 'messages.getById', ['form_params' => $this->robot->api->settings])->getBody()->getContents()));

        // Если в ответе ошибка
        if (isset($request->error)) {
            throw new Exception('Вконтакте: ' . $request->error->error_msg, $request->error->error_code);
        }

        return $request->response;
    }
}
